@@ -16,8 +16,121 @@
 #include <libxml++/io/ostreamoutputbuffer.h>
 
 #include <libxml/tree.h>
+#include <libxml/xinclude.h>
+#include <libxml/parser.h> // XML_PARSE_NOXINCNODE
 
 #include <iostream>
+#include <map>
+
+namespace // anonymous
+{
+typedef std::map<xmlpp::Node*, xmlElementType> NodeMap;
+
+// Find all C++ wrappers of 'node' and its descendants.
+// Compare xmlpp::Node::free_wrappers().
+void find_wrappers(xmlNode* node, NodeMap& node_map)
+{
+  if (!node)
+    return;
+    
+  //If an entity declaration contains an entity reference, there can be cyclic
+  //references between entity declarations and entity references. (It's not
+  //a tree.) We must avoid an infinite recursion.
+  //Compare xmlFreeNode(), which frees the children of all node types except
+  //XML_ENTITY_REF_NODE.
+  if (node->type != XML_ENTITY_REF_NODE)
+  {
+    // Walk the children list.
+    for (xmlNode* child = node->children; child; child = child->next)
+      find_wrappers(child, node_map);
+  }
+
+  // Find the local one
+  bool has_attributes = true;
+  switch (node->type)
+  {
+    // Node types that have no attributes.
+    // These are not represented by struct xmlNode.
+    case XML_DTD_NODE:
+    case XML_ATTRIBUTE_NODE:
+    case XML_ELEMENT_DECL:
+    case XML_ATTRIBUTE_DECL:
+    case XML_ENTITY_DECL:
+    case XML_DOCUMENT_NODE:
+      has_attributes = false;
+      break;
+    default:
+      break;
+  }
+
+  if (node->_private)
+    node_map[static_cast<xmlpp::Node*>(node->_private)] = node->type;
+
+  if (!has_attributes)
+    return;
+
+  //Walk the attributes list.
+  //Note that some "derived" structs have a different layout, so 
+  //_xmlNode::properties would be a nonsense value, leading to crashes
+  //(and shown as valgrind warnings), so we return above, to avoid 
+  //checking it here.
+  for (xmlAttr* attr = node->properties; attr; attr = attr->next)
+    find_wrappers(reinterpret_cast<xmlNode*>(attr), node_map);
+}
+
+// Remove from 'node_map' the pointers to the C++ wrappers that are found with
+// unchanged type in 'node' and its descendants.
+void remove_found_wrappers(xmlNode* node, NodeMap& node_map)
+{
+  if (!node)
+    return;
+    
+  if (node->type != XML_ENTITY_REF_NODE)
+  {
+    // Walk the children list.
+    for (xmlNode* child = node->children; child; child = child->next)
+      remove_found_wrappers(child, node_map);
+  }
+
+  // Find the local one
+  bool has_attributes = true;
+  switch (node->type)
+  {
+    // Node types that have no attributes
+    case XML_DTD_NODE:
+    case XML_ATTRIBUTE_NODE:
+    case XML_ELEMENT_DECL:
+    case XML_ATTRIBUTE_DECL:
+    case XML_ENTITY_DECL:
+    case XML_DOCUMENT_NODE:
+      has_attributes = false;
+      break;
+    default:
+      break;
+  }
+
+  if (node->_private)
+  {
+    const NodeMap::iterator iter =
+      node_map.find(static_cast<xmlpp::Node*>(node->_private));
+    if (iter != node_map.end())
+    {
+      if (iter->second == node->type)
+        node_map.erase(iter);
+      else
+        node->_private = 0; // node->type has changed. The wrapper will be deleted.
+    }
+  }
+
+  if (!has_attributes)
+    return;
+
+  // Walk the attributes list.
+  for (xmlAttr* attr = node->properties; attr; attr = attr->next)
+    remove_found_wrappers(reinterpret_cast<xmlNode*>(attr), node_map);
+
+}
+} // anonymous
 
 namespace xmlpp
 {
@@ -59,6 +172,8 @@ Document::Init Document::init_;
 Document::Document(const Glib::ustring& version)
   : impl_(xmlNewDoc((const xmlChar*)version.c_str()))
 {
+  if (!impl_)
+    throw internal_error("Could not create Document.");
   impl_->_private = this;
 }
 
@@ -125,11 +240,20 @@ Element* Document::create_root_node(const Glib::ustring& name,
                                     const Glib::ustring& ns_prefix)
 {
   xmlNode* node = xmlNewDocNode(impl_, 0, (const xmlChar*)name.c_str(), 0);
-  xmlDocSetRootElement(impl_, node);
+  if (!node)
+    throw internal_error("Could not create root element node " + name);
+
+  node = xmlDocSetRootElement(impl_, node);
+  if (node)
+  {
+    // An old root element node has been replaced.
+    Node::free_wrappers(node);
+    xmlFreeNode(node);
+  }
 
   Element* element = get_root_node();
 
-  if( !ns_uri.empty() )
+  if( !ns_uri.empty() && element )
   {
     element->set_namespace_declaration(ns_uri, ns_prefix);
     element->set_namespace(ns_prefix);
@@ -141,28 +265,38 @@ Element* Document::create_root_node(const Glib::ustring& name,
 Element* Document::create_root_node_by_import(const Node* node,
 					      bool recursive)
 {
+  if (!node)
+    return 0;
+
   //Create the node, by copying:
   xmlNode* imported_node = xmlDocCopyNode(const_cast<xmlNode*>(node->cobj()), impl_, recursive);
   if (!imported_node)
   {
-    throw exception("Unable to import node");
+    throw exception("Unable to copy the node that shall be imported");
   }
 
-  xmlDocSetRootElement(impl_, imported_node);
+  xmlNode* old_node = xmlDocSetRootElement(impl_, imported_node);
+  if (old_node)
+  {
+    // An old root element node has been replaced.
+    Node::free_wrappers(old_node);
+    xmlFreeNode(old_node);
+  }
 
   return get_root_node();
 }
 
 CommentNode* Document::add_comment(const Glib::ustring& content)
 {
-  xmlNode* node = xmlNewComment((const xmlChar*)content.c_str());
-  if(!node)
+  xmlNode* child = xmlNewComment((const xmlChar*)content.c_str());
+ 
+  // Use the result, because child can be freed when merging text nodes:
+  xmlNode* node = xmlAddChild((xmlNode*)impl_, child);
+  if (!node)
   {
-    throw internal_error("Cannot create comment node");
+    xmlFreeNode(child);
+    throw internal_error("Could not add comment node \"" + content + "\"");
   }
-
-  // Use the result, because node can be freed when merging text nodes:
-  node = xmlAddChild( (xmlNode*)impl_, node);
   Node::create_wrapper(node);
   return static_cast<CommentNode*>(node->_private);
 }
@@ -170,14 +304,15 @@ CommentNode* Document::add_comment(const Glib::ustring& content)
 ProcessingInstructionNode* Document::add_processing_instruction(
   const Glib::ustring& name, const Glib::ustring& content)
 {
-  xmlNode* node = xmlNewDocPI(impl_, (const xmlChar*)name.c_str(), (const xmlChar*)content.c_str());
-  if(!node)
+  xmlNode* child = xmlNewDocPI(impl_, (const xmlChar*)name.c_str(), (const xmlChar*)content.c_str());
+  xmlNode* node = xmlAddChild((xmlNode*)impl_, child);
+  if (!node)
   {
-    throw internal_error("Cannot create processing instruction node");
+    xmlFreeNode(child);
+    throw internal_error("Could not add processing instruction node " + name);
   }
-  node = xmlAddChild((xmlNode*)impl_, node);
   Node::create_wrapper(node);
-  return node ? static_cast<ProcessingInstructionNode*>(node->_private) : 0;
+  return static_cast<ProcessingInstructionNode*>(node->_private);
 }
 
 void Document::write_to_file(const Glib::ustring& filename, const Glib::ustring& encoding)
@@ -217,12 +352,13 @@ void Document::do_write_to_file(
 {
   KeepBlanks k(KeepBlanks::Default);
   xmlIndentTreeOutput = format?1:0;
+  xmlResetLastError();
   const int result = xmlSaveFormatFileEnc(filename.c_str(), impl_,
     get_encoding_or_utf8(encoding), format?1:0);
 
   if(result == -1)
   {
-    throw exception("do_write_to_file() failed.");
+    throw exception("do_write_to_file() failed.\n" + format_xml_error());
   }
 }
 
@@ -235,12 +371,13 @@ Glib::ustring Document::do_write_to_string(
   xmlChar* buffer = 0;
   int length = 0;
 
+  xmlResetLastError();
   xmlDocDumpFormatMemoryEnc(impl_, &buffer, &length,
     get_encoding_or_utf8(encoding), format?1:0);
 
   if(!buffer)
   {
-    throw exception("do_write_to_string() failed.");
+    throw exception("do_write_to_string() failed.\n" + format_xml_error());
   }
 
   // Create a Glib::ustring copy of the buffer
@@ -260,12 +397,13 @@ void Document::do_write_to_stream(std::ostream& output, const Glib::ustring& enc
 {
   // TODO assert document encoding is UTF-8 if encoding is different than UTF-8
   OStreamOutputBuffer buffer(output, encoding);
+  xmlResetLastError();
   const int result = xmlSaveFormatFileTo(buffer.cobj(), impl_,
     get_encoding_or_utf8(encoding), format ? 1 : 0);
   
   if(result == -1)
   {
-    throw exception("do_write_to_stream() failed.");
+    throw exception("do_write_to_stream() failed.\n" + format_xml_error());
   }
 }
 
@@ -273,10 +411,50 @@ void Document::set_entity_declaration(const Glib::ustring& name, XmlEntityType t
                               const Glib::ustring& publicId, const Glib::ustring& systemId,
                               const Glib::ustring& content)
 {
-  xmlAddDocEntity( impl_, (const xmlChar*) name.c_str(), type,
+  xmlEntity* entity = xmlAddDocEntity( impl_, (const xmlChar*) name.c_str(), type,
     publicId.empty() ? (const xmlChar*)0 : (const xmlChar*)publicId.c_str(),
     systemId.empty() ? (const xmlChar*)0 : (const xmlChar*)systemId.c_str(),
     (const xmlChar*) content.c_str() );
+  if (!entity)
+    throw internal_error("Could not add entity declaration " + name);
+}
+
+int Document::process_xinclude(bool generate_xinclude_nodes)
+{
+  NodeMap node_map;
+
+  xmlNode* root = xmlDocGetRootElement(impl_);
+
+  find_wrappers(root, node_map);
+
+  xmlResetLastError();
+  const int n_substitutions = xmlXIncludeProcessTreeFlags(root,
+    generate_xinclude_nodes ? 0 : XML_PARSE_NOXINCNODE);
+
+  remove_found_wrappers(reinterpret_cast<xmlNode*>(impl_), node_map);
+  // Delete wrappers of nodes that have been deleted or have got their type changed.
+  for (NodeMap::iterator iter = node_map.begin(); iter != node_map.end(); ++iter)
+  {
+    switch (iter->second)
+    {
+    case XML_DTD_NODE:
+      delete reinterpret_cast<Dtd*>(iter->first);
+      break;
+    case XML_DOCUMENT_NODE:
+      delete reinterpret_cast<Document*>(iter->first);
+      break;
+    default:
+      delete iter->first; // Node*
+      break;
+    }
+  }
+
+  if (n_substitutions < 0)
+  {
+    throw exception("Couldn't process XInclude\n" + format_xml_error());
+  }
+
+  return n_substitutions;
 }
 
 _xmlEntity* Document::get_entity(const Glib::ustring& name)
